@@ -23,7 +23,10 @@
 
 #include <sys/sysctl.h>
 #include <mach/machine.h>
-#include <CoreFoundation/CoreFoundation.h>
+
+#if defined(ARM32) && defined(ARM_THREAD_STATE32)
+# include <CoreFoundation/CoreFoundation.h>
+#endif
 
 /* From "Inside Mac OS X - Mach-O Runtime Architecture" published by Apple
    Page 49:
@@ -55,39 +58,48 @@ typedef struct StackFrame {
 
 GC_INNER ptr_t GC_FindTopOfStack(unsigned long stack_start)
 {
-  StackFrame *frame;
+  StackFrame *frame = (StackFrame *)stack_start;
 
-# ifdef POWERPC
-    if (stack_start == 0) {
+  if (stack_start == 0) {
+#   ifdef POWERPC
 #     if CPP_WORDSZ == 32
         __asm__ __volatile__ ("lwz %0,0(r1)" : "=r" (frame));
 #     else
         __asm__ __volatile__ ("ld %0,0(r1)" : "=r" (frame));
 #     endif
-    } else
-# else
-    GC_ASSERT(stack_start != 0); /* not implemented */
-# endif /* !POWERPC */
-  /* else */ {
-    frame = (StackFrame *)stack_start;
+#   elif defined(ARM32)
+        volatile ptr_t sp_reg;
+        __asm__ __volatile__ ("mov %0, r7\n" : "=r" (sp_reg));
+        frame = (StackFrame *)sp_reg;
+#   elif defined(AARCH64)
+        volatile ptr_t sp_reg;
+        __asm__ __volatile__ ("mov %0, x29\n" : "=r" (sp_reg));
+        frame = (StackFrame *)sp_reg;
+#   else
+#     if defined(CPPCHECK)
+        GC_noop1((word)&frame);
+#     endif
+      ABORT("GC_FindTopOfStack(0) is not implemented");
+#   endif
   }
 
 # ifdef DEBUG_THREADS_EXTRA
-    GC_log_printf("FindTopOfStack start at sp = %p\n", (void *)frame);
+    GC_log_printf("FindTopOfStack start at sp= %p\n", (void *)frame);
 # endif
-  while (frame->savedSP != 0) {
-    /* if there are no more stack frames, stop */
+  while (frame->savedSP != 0) { /* stop if no more stack frames */
+    unsigned long maskedLR;
 
     frame = (StackFrame*)frame->savedSP;
 
     /* we do these next two checks after going to the next frame
        because the LR for the first stack frame in the loop
        is not set up on purpose, so we shouldn't check it. */
-    if ((frame->savedLR & ~0x3) == 0 || (frame->savedLR & ~0x3) == ~0x3UL)
+    maskedLR = frame -> savedLR & ~0x3UL;
+    if (0 == maskedLR || ~0x3UL == maskedLR)
       break; /* if the next LR is bogus, stop */
   }
 # ifdef DEBUG_THREADS_EXTRA
-    GC_log_printf("FindTopOfStack finish at sp = %p\n", (void *)frame);
+    GC_log_printf("FindTopOfStack finish at sp= %p\n", (void *)frame);
 # endif
   return (ptr_t)frame;
 }
@@ -191,19 +203,21 @@ STATIC ptr_t GC_stack_range_for(ptr_t *phi, thread_act_t thread, GC_thread p,
     /* else */ {
       mach_msg_type_number_t thread_state_count = GC_MACH_THREAD_STATE_COUNT;
 
-      /* Get the thread state (registers, etc) */
-      kern_result = thread_get_state(thread, GC_MACH_THREAD_STATE,
-                                     (natural_t *)&state,
-                                     &thread_state_count);
+      /* Get the thread state (registers, etc.) */
+      do {
+        kern_result = thread_get_state(thread, GC_MACH_THREAD_STATE,
+                                       (natural_t *)&state,
+                                       &thread_state_count);
+      } while (kern_result == KERN_ABORTED);
     }
 #   ifdef DEBUG_THREADS
-      GC_log_printf("thread_get_state returns value = %d\n", kern_result);
+      GC_log_printf("thread_get_state returns %d\n", kern_result);
 #   endif
     if (kern_result != KERN_SUCCESS)
       ABORT("thread_get_state failed");
 
 #   if defined(I386)
-      lo = (void *)state.THREAD_FLD(esp);
+      lo = (ptr_t)state.THREAD_FLD(esp);
 #     ifndef DARWIN_DONT_PARSE_STACK
         *phi = GC_FindTopOfStack(state.THREAD_FLD(esp));
 #     endif
@@ -216,7 +230,7 @@ STATIC ptr_t GC_stack_range_for(ptr_t *phi, thread_act_t thread, GC_thread p,
       GC_push_one(state.THREAD_FLD(ebp));
 
 #   elif defined(X86_64)
-      lo = (void *)state.THREAD_FLD(rsp);
+      lo = (ptr_t)state.THREAD_FLD(rsp);
 #     ifndef DARWIN_DONT_PARSE_STACK
         *phi = GC_FindTopOfStack(state.THREAD_FLD(rsp));
 #     endif
@@ -227,7 +241,7 @@ STATIC ptr_t GC_stack_range_for(ptr_t *phi, thread_act_t thread, GC_thread p,
       GC_push_one(state.THREAD_FLD(rdi));
       GC_push_one(state.THREAD_FLD(rsi));
       GC_push_one(state.THREAD_FLD(rbp));
-      /* GC_push_one(state.THREAD_FLD(rsp)); */
+      /* rsp is skipped.        */
       GC_push_one(state.THREAD_FLD(r8));
       GC_push_one(state.THREAD_FLD(r9));
       GC_push_one(state.THREAD_FLD(r10));
@@ -238,11 +252,12 @@ STATIC ptr_t GC_stack_range_for(ptr_t *phi, thread_act_t thread, GC_thread p,
       GC_push_one(state.THREAD_FLD(r15));
 
 #   elif defined(POWERPC)
-      lo = (void *)(state.THREAD_FLD(r1) - PPC_RED_ZONE_SIZE);
+      lo = (ptr_t)(state.THREAD_FLD(r1) - PPC_RED_ZONE_SIZE);
 #     ifndef DARWIN_DONT_PARSE_STACK
         *phi = GC_FindTopOfStack(state.THREAD_FLD(r1));
 #     endif
       GC_push_one(state.THREAD_FLD(r0));
+      /* r1 is skipped. */
       GC_push_one(state.THREAD_FLD(r2));
       GC_push_one(state.THREAD_FLD(r3));
       GC_push_one(state.THREAD_FLD(r4));
@@ -275,24 +290,25 @@ STATIC ptr_t GC_stack_range_for(ptr_t *phi, thread_act_t thread, GC_thread p,
       GC_push_one(state.THREAD_FLD(r31));
 
 #   elif defined(ARM32)
-      lo = (void *)state.THREAD_FLD(sp);
+      lo = (ptr_t)state.THREAD_FLD(sp);
 #     ifndef DARWIN_DONT_PARSE_STACK
-        *phi = GC_FindTopOfStack(state.THREAD_FLD(sp));
+        *phi = GC_FindTopOfStack(state.THREAD_FLD(r[7])); /* fp */
 #     endif
       {
         int j;
-        for (j = 0; j <= 12; j++) {
+        for (j = 0; j < 7; j++)
+          GC_push_one(state.THREAD_FLD(r[j]));
+        j++; /* "r7" is skipped (iOS uses it as a frame pointer) */
+        for (; j <= 12; j++)
           GC_push_one(state.THREAD_FLD(r[j]));
-        }
       }
-      /* "pc" and "sp" are skipped */
+      /* "cpsr", "pc" and "sp" are skipped */
       GC_push_one(state.THREAD_FLD(lr));
-      GC_push_one(state.THREAD_FLD(cpsr));
 
 #   elif defined(AARCH64)
-      lo = (void *)state.THREAD_FLD(sp);
+      lo = (ptr_t)state.THREAD_FLD(sp);
 #     ifndef DARWIN_DONT_PARSE_STACK
-        *phi = GC_FindTopOfStack(state.THREAD_FLD(sp));
+        *phi = GC_FindTopOfStack(state.THREAD_FLD(fp));
 #     endif
       {
         int j;
@@ -300,14 +316,13 @@ STATIC ptr_t GC_stack_range_for(ptr_t *phi, thread_act_t thread, GC_thread p,
           GC_push_one(state.THREAD_FLD(x[j]));
         }
       }
-      /* "cpsr", "pc" and "sp" are skipped */
-      GC_push_one(state.THREAD_FLD(fp));
+      /* "cpsr", "fp", "pc" and "sp" are skipped */
       GC_push_one(state.THREAD_FLD(lr));
 
 #   elif defined(CPPCHECK)
       lo = NULL;
 #   else
-#     error FIXME for non-x86 || ppc || arm architectures
+#     error FIXME for non-arm/ppc/x86 architectures
 #   endif
   } /* thread != my_thread */
 
@@ -329,8 +344,12 @@ STATIC ptr_t GC_stack_range_for(ptr_t *phi, thread_act_t thread, GC_thread p,
   /* else */ {
     *paltstack_lo = NULL;
   }
+# if defined(STACKPTR_CORRECTOR_AVAILABLE) && defined(DARWIN_DONT_PARSE_STACK)
+    if (GC_sp_corrector != 0)
+      GC_sp_corrector((void **)&lo, (void *)(p -> id));
+# endif
 # ifdef DEBUG_THREADS
-    GC_log_printf("Darwin: Stack for thread %p = [%p,%p)\n",
+    GC_log_printf("Darwin: Stack for thread %p is [%p,%p)\n",
                   (void *)(word)thread, (void *)lo, (void *)(*phi));
 # endif
   return lo;
@@ -545,12 +564,13 @@ GC_INNER void GC_stop_world(void)
     GC_log_printf("Stopping the world from thread %p\n",
                   (void *)(word)my_thread);
 # endif
-# ifdef PARALLEL_MARK
-    if (GC_parallel) {
       /* Make sure all free list construction has stopped before we     */
       /* start.  No new construction can start, since free list         */
       /* construction is required to acquire and release the GC lock    */
       /* before it starts, and we have the lock.                        */
+
+# ifdef PARALLEL_MARK
+    if (GC_parallel) {
       GC_acquire_mark_lock();
       GC_ASSERT(GC_fl_builder_count == 0);
       /* We should have previously waited for it to become zero. */
@@ -631,7 +651,7 @@ GC_INNER void GC_stop_world(void)
   }
 
 # ifdef MPROTECT_VDB
-    if(GC_incremental) {
+    if (GC_auto_incremental) {
       GC_mprotect_stop();
     }
 # endif
@@ -653,6 +673,9 @@ GC_INLINE void GC_thread_resume(thread_act_t thread)
     struct thread_basic_info info;
     mach_msg_type_number_t outCount = THREAD_BASIC_INFO_COUNT;
 
+#   ifdef CPPCHECK
+      info.run_state = 0;
+#   endif
     kern_result = thread_info(thread, THREAD_BASIC_INFO,
                               (thread_info_t)&info, &outCount);
     if (kern_result != KERN_SUCCESS)
@@ -680,7 +703,7 @@ GC_INNER void GC_start_world(void)
     GC_log_printf("World starting\n");
 # endif
 # ifdef MPROTECT_VDB
-    if(GC_incremental) {
+    if (GC_auto_incremental) {
       GC_mprotect_resume();
     }
 # endif
